@@ -1,17 +1,20 @@
 # event/api.py
-# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: https://www.opensource.org/licenses/mit-license.php
 
-"""Public API functions for the event system.
+"""Public API functions for the event system."""
+from __future__ import annotations
 
-"""
-from __future__ import absolute_import
+from typing import Any
+from typing import Callable
 
 from .base import _registrars
+from .registry import _ET
 from .registry import _EventKey
+from .registry import _ListenerFnType
 from .. import exc
 from .. import util
 
@@ -20,9 +23,11 @@ CANCEL = util.symbol("CANCEL")
 NO_RETVAL = util.symbol("NO_RETVAL")
 
 
-def _event_key(target, identifier, fn):
+def _event_key(
+    target: _ET, identifier: str, fn: _ListenerFnType
+) -> _EventKey[_ET]:
     for evt_cls in _registrars[identifier]:
-        tgt = evt_cls._accept_with(target)
+        tgt = evt_cls._accept_with(target, identifier)
         if tgt is not None:
             return _EventKey(target, identifier, fn, tgt)
     else:
@@ -31,7 +36,9 @@ def _event_key(target, identifier, fn):
         )
 
 
-def listen(target, identifier, fn, *args, **kw):
+def listen(
+    target: Any, identifier: str, fn: Callable[..., Any], *args: Any, **kw: Any
+) -> None:
     """Register a listener function for the given target.
 
     The :func:`.listen` function is part of the primary interface for the
@@ -42,15 +49,14 @@ def listen(target, identifier, fn, *args, **kw):
         from sqlalchemy import event
         from sqlalchemy.schema import UniqueConstraint
 
+
         def unique_constraint_name(const, table):
-            const.name = "uq_%s_%s" % (
-                table.name,
-                list(const.columns)[0].name
-            )
+            const.name = "uq_%s_%s" % (table.name, list(const.columns)[0].name)
+
+
         event.listen(
-                UniqueConstraint,
-                "after_parent_attach",
-                unique_constraint_name)
+            UniqueConstraint, "after_parent_attach", unique_constraint_name
+        )
 
     :param bool insert: The default behavior for event handlers is to append
       the decorated user defined function to an internal list of registered
@@ -115,25 +121,25 @@ def listen(target, identifier, fn, *args, **kw):
     _event_key(target, identifier, fn).listen(*args, **kw)
 
 
-def listens_for(target, identifier, *args, **kw):
+def listens_for(
+    target: Any, identifier: str, *args: Any, **kw: Any
+) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
     """Decorate a function as a listener for the given target + identifier.
 
     The :func:`.listens_for` decorator is part of the primary interface for the
     SQLAlchemy event system, documented at :ref:`event_toplevel`.
 
-    This function generally shares the same kwargs as :func:`.listens`.
+    This function generally shares the same kwargs as :func:`.listen`.
 
     e.g.::
 
         from sqlalchemy import event
         from sqlalchemy.schema import UniqueConstraint
 
+
         @event.listens_for(UniqueConstraint, "after_parent_attach")
         def unique_constraint_name(const, table):
-            const.name = "uq_%s_%s" % (
-                table.name,
-                list(const.columns)[0].name
-            )
+            const.name = "uq_%s_%s" % (table.name, list(const.columns)[0].name)
 
     A given function can also be invoked for only the first invocation
     of the event using the ``once`` argument::
@@ -142,7 +148,6 @@ def listens_for(target, identifier, *args, **kw):
         def on_config():
             do_config()
 
-
     .. warning:: The ``once`` argument does not imply automatic de-registration
        of the listener function after it has been invoked a first time; a
        listener entry will remain associated with the target object.
@@ -156,14 +161,14 @@ def listens_for(target, identifier, *args, **kw):
 
     """
 
-    def decorate(fn):
+    def decorate(fn: Callable[..., Any]) -> Callable[..., Any]:
         listen(target, identifier, fn, *args, **kw)
         return fn
 
     return decorate
 
 
-def remove(target, identifier, fn):
+def remove(target: Any, identifier: str, fn: Callable[..., Any]) -> None:
     """Remove an event listener.
 
     The arguments here should match exactly those which were sent to
@@ -178,6 +183,7 @@ def remove(target, identifier, fn):
         def my_listener_function(*arg):
             pass
 
+
         # ... it's removed like this
         event.remove(SomeMappedClass, "before_insert", my_listener_function)
 
@@ -200,11 +206,6 @@ def remove(target, identifier, fn):
         events at high scale, use a mutable structure that is handled
         from inside of a single listener.
 
-        .. versionchanged:: 1.0.0 - a ``collections.deque()`` object is now
-           used as the container for the list of events, which explicitly
-           disallows collection mutation while the collection is being
-           iterated.
-
     .. seealso::
 
         :func:`.listen`
@@ -213,7 +214,7 @@ def remove(target, identifier, fn):
     _event_key(target, identifier, fn).remove()
 
 
-def contains(target, identifier, fn):
+def contains(target: Any, identifier: str, fn: Callable[..., Any]) -> bool:
     """Return True if the given target/ident/fn is set up to listen."""
 
     return _event_key(target, identifier, fn).contains()
