@@ -1,15 +1,18 @@
 # testing/assertions.py
-# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: https://www.opensource.org/licenses/mit-license.php
+# mypy: ignore-errors
 
-from __future__ import absolute_import
+
+from __future__ import annotations
 
 from collections import defaultdict
 import contextlib
 from copy import copy
+from itertools import filterfalse
 import re
 import sys
 import warnings
@@ -28,14 +31,13 @@ from .. import util
 from ..engine import default
 from ..engine import url
 from ..sql.selectable import LABEL_STYLE_TABLENAME_PLUS_COL
-from ..util import compat
 from ..util import decorator
 
 
 def expect_warnings(*messages, **kw):
     """Context manager which expects one or more warnings.
 
-    With no arguments, squelches all SAWarning and RemovedIn20Warning emitted via
+    With no arguments, squelches all SAWarning emitted via
     sqlalchemy.util.warn and sqlalchemy.util.warn_limited.   Otherwise
     pass string expressions that will match selected warnings via regex;
     all non-matching warnings are sent through.
@@ -45,9 +47,7 @@ def expect_warnings(*messages, **kw):
     Note that the test suite sets SAWarning warnings to raise exceptions.
 
     """  # noqa
-    return _expect_warnings(
-        (sa_exc.RemovedIn20Warning, sa_exc.SAWarning), messages, **kw
-    )
+    return _expect_warnings_sqla_only(sa_exc.SAWarning, messages, **kw)
 
 
 @contextlib.contextmanager
@@ -60,7 +60,7 @@ def expect_warnings_on(db, *messages, **kw):
     """
     spec = db_spec(db)
 
-    if isinstance(db, util.string_types) and not spec(config._current):
+    if isinstance(db, str) and not spec(config._current):
         yield
     else:
         with expect_warnings(*messages, **kw):
@@ -84,11 +84,15 @@ def emits_warning(*messages):
 
 
 def expect_deprecated(*messages, **kw):
-    return _expect_warnings(sa_exc.SADeprecationWarning, messages, **kw)
+    return _expect_warnings_sqla_only(
+        sa_exc.SADeprecationWarning, messages, **kw
+    )
 
 
 def expect_deprecated_20(*messages, **kw):
-    return _expect_warnings(sa_exc.Base20DeprecationWarning, messages, **kw)
+    return _expect_warnings_sqla_only(
+        sa_exc.Base20DeprecationWarning, messages, **kw
+    )
 
 
 def emits_warning_on(db, *messages):
@@ -140,6 +144,29 @@ _SEEN = None
 _EXC_CLS = None
 
 
+def _expect_warnings_sqla_only(
+    exc_cls,
+    messages,
+    regex=True,
+    search_msg=False,
+    assert_=True,
+):
+    """SQLAlchemy internal use only _expect_warnings().
+
+    Alembic is using _expect_warnings() directly, and should be updated
+    to use this new interface.
+
+    """
+    return _expect_warnings(
+        exc_cls,
+        messages,
+        regex=regex,
+        search_msg=search_msg,
+        assert_=assert_,
+        raise_on_any_unexpected=True,
+    )
+
+
 @contextlib.contextmanager
 def _expect_warnings(
     exc_cls,
@@ -147,11 +174,9 @@ def _expect_warnings(
     regex=True,
     search_msg=False,
     assert_=True,
-    py2konly=False,
     raise_on_any_unexpected=False,
     squelch_other_warnings=False,
 ):
-
     global _FILTERS, _SEEN, _EXC_CLS
 
     if regex or search_msg:
@@ -182,7 +207,6 @@ def _expect_warnings(
             real_warn = warnings.warn
 
         def our_warn(msg, *arg, **kw):
-
             if isinstance(msg, _EXC_CLS):
                 exception = type(msg)
                 msg = str(msg)
@@ -212,19 +236,13 @@ def _expect_warnings(
                 if not squelch_other_warnings:
                     real_warn(msg, *arg, **kw)
 
-        with mock.patch("warnings.warn", our_warn), mock.patch(
-            "sqlalchemy.util.SQLALCHEMY_WARN_20", True
-        ), mock.patch(
-            "sqlalchemy.util.deprecations.SQLALCHEMY_WARN_20", True
-        ), mock.patch(
-            "sqlalchemy.engine.row.LegacyRow._default_key_style", 2
-        ):
+        with mock.patch("warnings.warn", our_warn):
             try:
                 yield
             finally:
                 _SEEN = _FILTERS = _EXC_CLS = None
 
-                if assert_ and (not py2konly or not compat.py3k):
+                if assert_:
                     assert not seen, "Warnings were not seen: %s" % ", ".join(
                         "%r" % (s.pattern if regex else s) for s in seen
                     )
@@ -256,8 +274,8 @@ def int_within_variance(expected, received, variance):
     )
 
 
-def eq_regex(a, b, msg=None):
-    assert re.match(b, a), msg or "%r !~ %r" % (a, b)
+def eq_regex(a, b, msg=None, flags=0):
+    assert re.match(b, a, flags), msg or "%r !~ %r" % (a, b)
 
 
 def eq_(a, b, msg=None):
@@ -352,9 +370,6 @@ def _assert_proper_exception_context(exception):
 
     """
 
-    if not util.py3k:
-        return
-
     if (
         exception.__context__ is not exception.__cause__
         and not exception.__suppress_context__
@@ -389,7 +404,7 @@ def assert_warns(except_cls, callable_, *args, **kwargs):
 
 
     """
-    with _expect_warnings(except_cls, [".*"], squelch_other_warnings=True):
+    with _expect_warnings_sqla_only(except_cls, [".*"]):
         return callable_(*args, **kwargs)
 
 
@@ -404,12 +419,11 @@ def assert_warns_message(except_cls, msg, callable_, *args, **kwargs):
     rather than regex.match().
 
     """
-    with _expect_warnings(
+    with _expect_warnings_sqla_only(
         except_cls,
         [msg],
         search_msg=True,
         regex=False,
-        squelch_other_warnings=True,
     ):
         return callable_(*args, **kwargs)
 
@@ -423,13 +437,12 @@ def assert_raises_message_context_ok(
 def _assert_raises(
     except_cls, callable_, args, kwargs, msg=None, check_context=False
 ):
-
     with _expect_raises(except_cls, msg, check_context) as ec:
         callable_(*args, **kwargs)
     return ec.error
 
 
-class _ErrorContainer(object):
+class _ErrorContainer:
     error = None
 
 
@@ -454,12 +467,16 @@ def _expect_raises(except_cls, msg=None, check_context=False):
         ec.error = err
         success = True
         if msg is not None:
-            assert re.search(
-                msg, util.text_type(err), re.UNICODE
-            ), "%r !~ %s" % (msg, err)
+            # I'm often pdbing here, and "err" above isn't
+            # in scope, so assign the string explicitly
+            error_as_string = str(err)
+            assert re.search(msg, error_as_string, re.UNICODE), "%r !~ %s" % (
+                msg,
+                error_as_string,
+            )
         if check_context and not are_we_already_in_a_traceback:
             _assert_proper_exception_context(err)
-        print(util.text_type(err).encode("utf-8"))
+        print(str(err).encode("utf-8"))
 
     # it's generally a good idea to not carry traceback objects outside
     # of the except: block, but in this case especially we seem to have
@@ -480,7 +497,7 @@ def expect_raises_message(except_cls, msg, check_context=True):
     return _expect_raises(except_cls, msg=msg, check_context=check_context)
 
 
-class AssertsCompiledSQL(object):
+class AssertsCompiledSQL:
     def assert_compile(
         self,
         clause,
@@ -496,6 +513,7 @@ class AssertsCompiledSQL(object):
         use_default_dialect=False,
         allow_dialect_select=False,
         supports_default_values=True,
+        supports_native_boolean=False,
         supports_default_metavalue=True,
         literal_binds=False,
         render_postcompile=False,
@@ -504,11 +522,13 @@ class AssertsCompiledSQL(object):
         default_schema_name=None,
         from_linting=False,
         check_param_order=True,
+        use_literal_execute_for_simple_int=False,
     ):
         if use_default_dialect:
             dialect = default.DefaultDialect()
             dialect.supports_default_values = supports_default_values
             dialect.supports_default_metavalue = supports_default_metavalue
+            dialect.supports_native_boolean = supports_native_boolean
         elif allow_dialect_select:
             dialect = None
         else:
@@ -521,12 +541,12 @@ class AssertsCompiledSQL(object):
                 if dialect == "default":
                     dialect = default.DefaultDialect()
                 else:
-                    dialect = default.DefaultDialect(paramstyle="qmark")
+                    dialect = default.DefaultDialect("qmark")
                 dialect.supports_default_values = supports_default_values
                 dialect.supports_default_metavalue = supports_default_metavalue
             elif dialect == "default_enhanced":
                 dialect = default.StrCompileDialect()
-            elif isinstance(dialect, util.string_types):
+            elif isinstance(dialect, str):
                 dialect = url.URL.create(dialect).get_dialect()()
 
         if default_schema_name:
@@ -547,6 +567,9 @@ class AssertsCompiledSQL(object):
         if render_postcompile:
             compile_kwargs["render_postcompile"] = True
 
+        if use_literal_execute_for_simple_int:
+            compile_kwargs["use_literal_execute_for_simple_int"] = True
+
         if for_executemany:
             kw["for_executemany"] = True
 
@@ -566,14 +589,14 @@ class AssertsCompiledSQL(object):
         if compile_kwargs:
             kw["compile_kwargs"] = compile_kwargs
 
-        class DontAccess(object):
+        class DontAccess:
             def __getattribute__(self, key):
                 raise NotImplementedError(
                     "compiler accessed .statement; use "
                     "compiler.current_executable"
                 )
 
-        class CheckCompilerAccess(object):
+        class CheckCompilerAccess:
             def __init__(self, test_statement):
                 self.test_statement = test_statement
                 self._annotations = {}
@@ -591,6 +614,10 @@ class AssertsCompiledSQL(object):
                     if hasattr(test_statement, "_return_defaults"):
                         self._return_defaults = test_statement._return_defaults
 
+            @property
+            def _variant_mapping(self):
+                return self.test_statement._variant_mapping
+
             def _default_dialect(self):
                 return self.test_statement._default_dialect()
 
@@ -632,29 +659,38 @@ class AssertsCompiledSQL(object):
                 hash(cache_key_no_warnings[0])
 
         param_str = repr(getattr(c, "params", {}))
-        if util.py3k:
-            param_str = param_str.encode("utf-8").decode("ascii", "ignore")
-            print(
-                ("\nSQL String:\n" + util.text_type(c) + param_str).encode(
-                    "utf-8"
-                )
-            )
-        else:
-            print(
-                "\nSQL String:\n"
-                + util.text_type(c).encode("utf-8")
-                + param_str
-            )
+        param_str = param_str.encode("utf-8").decode("ascii", "ignore")
+        print(("\nSQL String:\n" + str(c) + param_str).encode("utf-8"))
 
-        cc = re.sub(r"[\n\t]", "", util.text_type(c))
+        cc = re.sub(r"[\n\t]", "", str(c))
 
         eq_(cc, result, "%r != %r on dialect %r" % (cc, result, dialect))
 
         if checkparams is not None:
-            eq_(c.construct_params(params), checkparams)
+            if render_postcompile:
+                expanded_state = c.construct_expanded_state(
+                    params, escape_names=False
+                )
+                eq_(expanded_state.parameters, checkparams)
+            else:
+                eq_(c.construct_params(params), checkparams)
         if checkpositional is not None:
-            p = c.construct_params(params, escape_names=False)
-            eq_(tuple([p[x] for x in c.positiontup]), checkpositional)
+            if render_postcompile:
+                expanded_state = c.construct_expanded_state(
+                    params, escape_names=False
+                )
+                eq_(
+                    tuple(
+                        [
+                            expanded_state.parameters[x]
+                            for x in expanded_state.positiontup
+                        ]
+                    ),
+                    checkpositional,
+                )
+            else:
+                p = c.construct_params(params, escape_names=False)
+                eq_(tuple([p[x] for x in c.positiontup]), checkpositional)
         if check_prefetch is not None:
             eq_(c.prefetch, check_prefetch)
         if check_literal_execute is not None:
@@ -683,7 +719,7 @@ class AssertsCompiledSQL(object):
 
             pyformat_dialect = get_dialect("pyformat", False)
             pyformat_c = clause.compile(dialect=pyformat_dialect, **kw)
-            stmt = re.sub(r"[\n\t]", "", pyformat_c.string)
+            stmt = re.sub(r"[\n\t]", "", str(pyformat_c))
 
             qmark_dialect = get_dialect("qmark", True)
             qmark_c = clause.compile(dialect=qmark_dialect, **kw)
@@ -727,14 +763,22 @@ class AssertsCompiledSQL(object):
             is_true(ordered, msg)
 
 
-class ComparesTables(object):
-    def assert_tables_equal(self, table, reflected_table, strict_types=False):
+class ComparesTables:
+    def assert_tables_equal(
+        self,
+        table,
+        reflected_table,
+        strict_types=False,
+        strict_constraints=True,
+    ):
         assert len(table.c) == len(reflected_table.c)
         for c, reflected_c in zip(table.c, reflected_table.c):
             eq_(c.name, reflected_c.name)
             assert reflected_c is reflected_table.c[c.name]
-            eq_(c.primary_key, reflected_c.primary_key)
-            eq_(c.nullable, reflected_c.nullable)
+
+            if strict_constraints:
+                eq_(c.primary_key, reflected_c.primary_key)
+                eq_(c.nullable, reflected_c.nullable)
 
             if strict_types:
                 msg = "Type '%s' doesn't correspond to type '%s'"
@@ -748,18 +792,20 @@ class ComparesTables(object):
             if isinstance(c.type, sqltypes.String):
                 eq_(c.type.length, reflected_c.type.length)
 
-            eq_(
-                {f.column.name for f in c.foreign_keys},
-                {f.column.name for f in reflected_c.foreign_keys},
-            )
+            if strict_constraints:
+                eq_(
+                    {f.column.name for f in c.foreign_keys},
+                    {f.column.name for f in reflected_c.foreign_keys},
+                )
             if c.server_default:
                 assert isinstance(
                     reflected_c.server_default, schema.FetchedValue
                 )
 
-        assert len(table.primary_key) == len(reflected_table.primary_key)
-        for c in table.primary_key:
-            assert reflected_table.primary_key.columns[c.name] is not None
+        if strict_constraints:
+            assert len(table.primary_key) == len(reflected_table.primary_key)
+            for c in table.primary_key:
+                assert reflected_table.primary_key.columns[c.name] is not None
 
     def assert_types_base(self, c1, c2):
         assert c1.type._compare_type_affinity(
@@ -771,7 +817,7 @@ class ComparesTables(object):
         )
 
 
-class AssertsExecutionResults(object):
+class AssertsExecutionResults:
     def assert_result(self, result, class_, *objects):
         result = list(result)
         print(repr(result))
@@ -818,9 +864,7 @@ class AssertsExecutionResults(object):
         found = util.IdentitySet(result)
         expected = {immutabledict(e) for e in expected}
 
-        for wrong in util.itertools_filterfalse(
-            lambda o: isinstance(o, cls), found
-        ):
+        for wrong in filterfalse(lambda o: isinstance(o, cls), found):
             fail(
                 'Unexpected type "%s", expected "%s"'
                 % (type(wrong).__name__, cls.__name__)
@@ -873,7 +917,6 @@ class AssertsExecutionResults(object):
         return result
 
     def assert_sql(self, db, callable_, rules):
-
         newrules = []
         for rule in rules:
             if isinstance(rule, dict):
@@ -887,25 +930,10 @@ class AssertsExecutionResults(object):
         return self.assert_sql_execution(db, callable_, *newrules)
 
     def assert_sql_count(self, db, callable_, count):
-        self.assert_sql_execution(
+        return self.assert_sql_execution(
             db, callable_, assertsql.CountStatements(count)
         )
 
-    def assert_multiple_sql_count(self, dbs, callable_, counts):
-        recs = [
-            (self.sql_execution_asserter(db), db, count)
-            for (db, count) in zip(dbs, counts)
-        ]
-        asserters = []
-        for ctx, db, count in recs:
-            asserters.append(ctx.__enter__())
-        try:
-            return callable_()
-        finally:
-            for asserter, (ctx, db, count) in zip(asserters, recs):
-                ctx.__exit__(None, None, None)
-                asserter.assert_(assertsql.CountStatements(count))
-
     @contextlib.contextmanager
     def assert_execution(self, db, *rules):
         with self.sql_execution_asserter(db) as asserter:
@@ -914,3 +942,50 @@ class AssertsExecutionResults(object):
 
     def assert_statement_count(self, db, count):
         return self.assert_execution(db, assertsql.CountStatements(count))
+
+    @contextlib.contextmanager
+    def assert_statement_count_multi_db(self, dbs, counts):
+        recs = [
+            (self.sql_execution_asserter(db), db, count)
+            for (db, count) in zip(dbs, counts)
+        ]
+        asserters = []
+        for ctx, db, count in recs:
+            asserters.append(ctx.__enter__())
+        try:
+            yield
+        finally:
+            for asserter, (ctx, db, count) in zip(asserters, recs):
+                ctx.__exit__(None, None, None)
+                asserter.assert_(assertsql.CountStatements(count))
+
+
+class ComparesIndexes:
+    def compare_table_index_with_expected(
+        self, table: schema.Table, expected: list, dialect_name: str
+    ):
+        eq_(len(table.indexes), len(expected))
+        idx_dict = {idx.name: idx for idx in table.indexes}
+        for exp in expected:
+            idx = idx_dict[exp["name"]]
+            eq_(idx.unique, exp["unique"])
+            cols = [c for c in exp["column_names"] if c is not None]
+            eq_(len(idx.columns), len(cols))
+            for c in cols:
+                is_true(c in idx.columns)
+            exprs = exp.get("expressions")
+            if exprs:
+                eq_(len(idx.expressions), len(exprs))
+                for idx_exp, expr, col in zip(
+                    idx.expressions, exprs, exp["column_names"]
+                ):
+                    if col is None:
+                        eq_(idx_exp.text, expr)
+            if (
+                exp.get("dialect_options")
+                and f"{dialect_name}_include" in exp["dialect_options"]
+            ):
+                eq_(
+                    idx.dialect_options[dialect_name]["include"],
+                    exp["dialect_options"][f"{dialect_name}_include"],
+                )
