@@ -315,6 +315,13 @@ class SuiteRequirements(Requirements):
 
         return exclusions.closed()
 
+    @property
+    def ctes_with_values(self):
+        """target database supports CTES that ride on top of a VALUES
+        clause."""
+
+        return exclusions.closed()
+
     @property
     def ctes_on_dml(self):
         """target database supports CTES which consist of INSERT, UPDATE
@@ -658,6 +665,12 @@ class SuiteRequirements(Requirements):
 
         return exclusions.closed()
 
+    @property
+    def temp_table_comment_reflection(self):
+        """indicates if database supports comments on temp tables and
+        the dialect can reflect them"""
+        return exclusions.closed()
+
     @property
     def comment_reflection(self):
         """Indicates if the database support table comment reflection"""
@@ -823,6 +836,11 @@ class SuiteRequirements(Requirements):
 
         return exclusions.open()
 
+    @property
+    def nvarchar_types(self):
+        """target database supports NVARCHAR and NCHAR as an actual datatype"""
+        return exclusions.closed()
+
     @property
     def unicode_data_no_special_types(self):
         """Target database/dialect can receive / deliver / compare data with
